from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

from ..module_utils.my_util import question  # pylint: disable=unused-import


def action_code():
    return "hello from my_action.py"
