#!/usr/bin/python
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from ansible.module_utils.basic import AnsibleModule


if __name__ == '__main__':
    module = AnsibleModule(argument_spec=dict(
        test=dict(type='str', choices=['foo', 'bar'], default='foo'),
    ))
    module.exit_json(test='foo')
