[windows]
{% for host in vars.groups.windows %}
{{ host }}  ansible_host={{ hostvars[host]['ansible_host'] }} ansible_port={{ hostvars[host]['ansible_port'] }} ansible_user={{ hostvars[host]['ansible_user'] }} ansible_password={{ hostvars[host]['ansible_password'] }}
{% endfor %}

[windows:vars]
ansible_connection=psrp
ansible_psrp_auth=negotiate
ansible_psrp_cert_validation=ignore
